"""Collection of helper functions for assessing and performing automated data transfers related to AWS."""

import json
import os
import time
from datetime import datetime
from typing import Optional
from uuid import uuid4


def submit_aws_batch_job(
    *,
    job_name: str,
    docker_image: str,
    commands: Optional[list[str]] = None,
    environment_variables: Optional[dict[str, str]] = None,
    efs_volume_name: Optional[str] = None,
    job_dependencies: Optional[list[dict[str, str]]] = None,
    status_tracker_table_name: str = "neuroconv_batch_status_tracker",
    iam_role_name: str = "neuroconv_batch_role",
    compute_environment_name: str = "neuroconv_batch_environment",
    job_queue_name: str = "neuroconv_batch_queue",
    job_definition_name: Optional[str] = None,
    minimum_worker_ram_in_gib: int = 4,
    minimum_worker_cpus: int = 4,
    submission_id: Optional[str] = None,
    region: Optional[str] = None,
) -> dict[str, str]:
    """
    Submit a job to AWS Batch for processing.

    Requires AWS credentials saved to files in the `~/.aws/` folder or set as environment variables.

    Parameters
    ----------
    job_name : str
        The name of the job to submit.
    docker_image : str
        The name of the Docker image to use for the job.
    commands : str, optional
        The list of commands to run in the Docker container. Normal spaces are separate entries in the list.
        Current syntax only supports a single line; consecutive actions should be chained with the '&&' operator.
        E.g., `commands=["echo", "'Hello, World!'"]`.
    environment_variables : dict, optional
        A dictionary of environment variables to pass to the Docker container.
    efs_volume_name : str, optional
        The name of an EFS volume to be created and attached to the job.
        The path exposed to the container will always be `/mnt/efs`.
    job_dependencies : list of dict
        A list of job dependencies for this job to trigger. Structured as follows:
        [
            {"jobId": "job_id_1", "type": "N_TO_N"},
            {"jobId": "job_id_2", "type": "SEQUENTIAL"},
            ...
        ]

        Refer to the boto3 API documentation for latest syntax.
    status_tracker_table_name : str, default: "neuroconv_batch_status_tracker"
        The name of the DynamoDB table to use for tracking job status.
    iam_role_name : str, default: "neuroconv_batch_role"
        The name of the IAM role to use for the job.
    compute_environment_name : str, default: "neuroconv_batch_environment"
        The name of the compute environment to use for the job.
    job_queue_name : str, default: "neuroconv_batch_queue"
        The name of the job queue to use for the job.
    job_definition_name : str, optional
        The name of the job definition to use for the job.
        If unspecified, a name starting with 'neuroconv_batch_' will be generated.
    minimum_worker_ram_in_gib : int, default: 4
        The minimum amount of base worker memory required to run this job.
        Determines the EC2 instance type selected by the automatic 'best fit' selector.
        Recommended to be several GiB to allow comfortable buffer space for data chunk iterators.
    minimum_worker_cpus : int, default: 4
        The minimum number of CPUs required to run this job.
        A minimum of 4 is required, even if only one will be used in the actual process.
    submission_id : str, optional
        The unique ID to pair with this job submission when tracking the status via DynamoDB.
        Defaults to a random UUID4.
    region : str, optional
        The AWS region to use for the job.
        If not provided, we will attempt to load the region from your local AWS configuration.
        If that file is not found on your system, we will default to "us-east-2", the location of the DANDI Archive.

    Returns
    -------
    info : dict
        A dictionary containing information about this AWS Batch job.

        info["job_submission_info"] is the return value of `boto3.client.submit_job` which contains the job ID.
        info["table_submission_info"] is the initial row data inserted into the DynamoDB status tracking table.
    """
    import boto3

    region = region or "us-east-2"
    subregion = region + "a"  # For anything that requires subregion, always default to "a"
    aws_access_key_id = os.environ.get("AWS_ACCESS_KEY_ID", None)
    aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY", None)

    # Initialize all clients
    dynamodb_client = boto3.client(
        service_name="dynamodb",
        region_name=region,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )
    dynamodb_resource = boto3.resource(
        service_name="dynamodb",
        region_name=region,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )
    iam_client = boto3.client(
        service_name="iam",
        region_name=region,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )
    batch_client = boto3.client(
        service_name="batch",
        region_name=region,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )
    efs_client = boto3.client(
        service_name="efs",
        region_name=region,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )

    # Get the tracking table and IAM role
    table = _create_or_get_status_tracker_table(
        status_tracker_table_name=status_tracker_table_name,
        dynamodb_client=dynamodb_client,
        dynamodb_resource=dynamodb_resource,
    )
    iam_role_info = _create_or_get_iam_role(iam_role_name=iam_role_name, iam_client=iam_client)

    # Ensure all job submission requirements are met
    _ensure_compute_environment_exists(compute_environment_name=compute_environment_name, batch_client=batch_client)
    _ensure_job_queue_exists(
        job_queue_name=job_queue_name, compute_environment_name=compute_environment_name, batch_client=batch_client
    )

    efs_id = _create_or_get_efs_id(efs_volume_name=efs_volume_name, efs_client=efs_client, region=region)
    job_definition_name = job_definition_name or _generate_job_definition_name(
        docker_image=docker_image,
        minimum_worker_ram_in_gib=minimum_worker_ram_in_gib,
        minimum_worker_cpus=minimum_worker_cpus,
        efs_id=efs_id,
    )
    job_definition_arn = _ensure_job_definition_exists_and_get_arn(
        job_definition_name=job_definition_name,
        docker_image=docker_image,
        minimum_worker_ram_in_gib=minimum_worker_ram_in_gib,
        minimum_worker_cpus=minimum_worker_cpus,
        role_info=iam_role_info,
        batch_client=batch_client,
        efs_id=efs_id,
    )

    # Submit job and update status tracker
    currently_running_jobs = batch_client.list_jobs(jobQueue=job_queue_name)
    if job_name in currently_running_jobs:
        raise ValueError(
            f"There is already a job named '{job_name}' running in the queue! "
            "If you are submitting multiple jobs, each will need a unique name."
        )

    # Set environment variables to the docker container as well as optional commands to run
    job_dependencies = job_dependencies or []
    container_overrides = dict()
    if environment_variables is not None:
        container_overrides["environment"] = [{key: value} for key, value in environment_variables.items()]
    if commands is not None:
        container_overrides["command"] = commands

    job_submission_info = batch_client.submit_job(
        jobName=job_name,
        dependsOn=job_dependencies,
        jobDefinition=job_definition_arn,
        jobQueue=job_queue_name,
        containerOverrides=container_overrides,
    )

    # Update status tracking table
    submission_id = submission_id or str(uuid4())
    table_submission_info = dict(
        id=submission_id,
        job_id=job_submission_info["jobId"],
        job_name=job_name,
        submitted_on=datetime.now().isoformat(),
        status="Job submitted...",
    )
    table.put_item(Item=table_submission_info)

    info = dict(job_submission_info=job_submission_info, table_submission_info=table_submission_info)

    return info


def _create_or_get_status_tracker_table(
    *,
    status_tracker_table_name: str,
    dynamodb_client: "boto3.client.dynamodb",
    dynamodb_resource: "boto3.resources.dynamodb",
) -> "boto3.resources.dynamodb.Table":  # pragma: no cover
    """
    Create or get the DynamoDB table for tracking the status of jobs submitted to AWS Batch.

    It is very useful to have a status tracker that is separate from the job environment.

    Detailed logs of inner workings of each instance are given in the CloudWatch, but that can only really be
    analyzed from the AWS web console.

    Parameters
    ----------
    status_tracker_table_name : str, default: "neuroconv_batch_status_tracker"
        The name of the DynamoDB table to use for tracking job status.
    dynamodb_client : boto3.client.dynamodb
        The DynamoDB client to use for the job.
    dynamodb_resource : boto3.resources.dynamodb
        The DynamoDB resource to use for the job.
    """
    current_tables = dynamodb_client.list_tables()["TableNames"]
    if status_tracker_table_name not in current_tables:
        table = dynamodb_resource.create_table(
            TableName=status_tracker_table_name,
            KeySchema=[dict(AttributeName="id", KeyType="HASH")],
            AttributeDefinitions=[dict(AttributeName="id", AttributeType="S")],
            ProvisionedThroughput=dict(ReadCapacityUnits=1, WriteCapacityUnits=1),
        )
    else:
        table = dynamodb_resource.Table(name=status_tracker_table_name)

    return table


def _create_or_get_iam_role(*, iam_role_name: str, iam_client: "boto3.client.iam") -> dict:  # pragma: no cover
    """
    Create or get the IAM role policy for the AWS Batch job.

    Parameters
    ----------
    iam_role_name : str
        The name of the IAM role to use for the job.
    iam_client : boto3.client.iam
        The IAM client to use for the job.

    Returns
    -------
    role_info : dict
        All associated information about the IAM role (including set policies).
    """
    # Normally try/except structures are not recommended
    # But making a targeted request for a named entity here incurs less AWS costs than listing all entities
    try:
        role_info = iam_client.get_role(RoleName=iam_role_name)
        return role_info
    except Exception as exception:
        if "NoSuchEntity" not in str(exception):
            raise exception

    assume_role_policy = dict(
        Version="2012-10-17",
        Statement=[
            dict(Effect="Allow", Principal=dict(Service="ecs-tasks.amazonaws.com"), Action="sts:AssumeRole"),
        ],
    )

    # iam_client.create_role() is synchronous and so there is no need to wait in this function
    role_info = iam_client.create_role(RoleName=iam_role_name, AssumeRolePolicyDocument=json.dumps(assume_role_policy))
    iam_client.attach_role_policy(
        RoleName=role_info["Role"]["RoleName"], PolicyArn="arn:aws:iam::aws:policy/AmazonDynamoDBFullAccess"
    )
    iam_client.attach_role_policy(
        RoleName=role_info["Role"]["RoleName"], PolicyArn="arn:aws:iam::aws:policy/CloudWatchFullAccess"
    )

    return role_info


def _ensure_compute_environment_exists(
    *, compute_environment_name: str, batch_client: "boto3.client.Batch", max_retries: int = 12
) -> None:  # pragma: no cover
    """
    Ensure that the compute environment exists in AWS Batch.

    Parameters
    ----------
    compute_environment_name : str
        The name of the compute environment to use for the job.
    batch_client : boto3.client.Batch
        The AWS Batch client to use for the job.
    max_retries : int, default: 12
        If the compute environment does not already exist, then this is the maximum number of times to synchronously
        check for its successful creation before erroring.
        This is essential for a clean setup of the entire pipeline, or else later steps might error because they tried
        to launch before the compute environment was ready.
    """
    compute_environment_request = batch_client.describe_compute_environments(
        computeEnvironments=[compute_environment_name]
    )
    compute_environment_response = compute_environment_request["computeEnvironments"]

    if len(compute_environment_response) == 1:
        return compute_environment_response[0]

    if len(compute_environment_response) > 1:
        raise ValueError(
            f"Multiple compute environments with the name '{compute_environment_name}' were found! "
            "Please ensure that only one compute environment has this name."
        )

    # batch_client.create_compute_environment() is not synchronous and so we need to wait a bit afterwards
    batch_client.create_compute_environment(
        computeEnvironmentName=compute_environment_name,
        type="MANAGED",
        state="ENABLED",
        computeResources={
            "type": "EC2",
            "allocationStrategy": "BEST_FIT",  # Note: not currently supporting spot due to interruptibility
            "instanceTypes": ["optimal"],
            "minvCpus": 0,  # Note: if not zero, will always keep an instance running in active state on standby
            "maxvCpus": 8,  # Note: not currently exposing control over this since these are mostly I/O intensive
            "instanceRole": "ecsInstanceRole",
            # Security groups and subnets last updated on 8/4/2024
            "securityGroupIds": ["sg-001699e5b7496b226"],
            "subnets": [
                "subnet-0890a93aedb42e73e",  # us-east-2a
                "subnet-0e20bbcfb951b5387",  # us-east-2b
                "subnet-0680e07980538b786",  # us-east-2c
            ],
        },
    )

    compute_environment_request = batch_client.describe_compute_environments(
        computeEnvironments=[compute_environment_name]
    )
    compute_environment_response = compute_environment_request["computeEnvironments"]
    compute_environment_status = (
        compute_environment_response[0]["status"] if len(compute_environment_response) == 1 else ""
    )
    retry_count = 0
    while compute_environment_status != "VALID" and retry_count <= max_retries:
        retry_count += 1
        time.sleep(10)
        compute_environment_request = batch_client.describe_compute_environments(
            computeEnvironments=[compute_environment_name]
        )
        compute_environment_response = compute_environment_request["computeEnvironments"]
        compute_environment_status = (
            compute_environment_response[0]["status"] if len(compute_environment_response) == 1 else ""
        )

    if compute_environment_status != "VALID":
        raise ValueError(
            f"Compute environment '{compute_environment_name}' failed to launch after {max_retries} retries."
        )

    return None


def _ensure_job_queue_exists(
    *, job_queue_name: str, compute_environment_name: str, batch_client: "boto3.client.Batch", max_retries: int = 12
) -> None:  # pragma: no cover
    """
    Ensure that the job queue exists in AWS Batch.

    Parameters
    ----------
    job_queue_name : str
        The name of the job queue to use for the job.
    compute_environment_name : str
        The name of the compute environment to associate with the job queue.
    batch_client : boto3.client.Batch
        The AWS Batch client to use for the job.
    max_retries : int, default: 12
        If the job queue does not already exist, then this is the maximum number of times to synchronously
        check for its successful creation before erroring.
        This is essential for a clean setup of the entire pipeline, or else later steps might error because they tried
        to launch before the job queue was ready.
    """
    job_queue_request = batch_client.describe_job_queues(jobQueues=[job_queue_name])
    job_queue_response = job_queue_request["jobQueues"]

    if len(job_queue_response) == 1:
        return job_queue_response[0]

    if len(job_queue_response) > 1:
        raise ValueError(
            f"Multiple job queues with the name '{job_queue_name}' were found! "
            "Please ensure that only one job queue has this name."
        )

    # Note: jobs submitted to EC2 Batch can occasionally get stuck in 'runnable' status due to bad configuration of
    # worker memory/CPU or general resource contention.
    # Eventually consider exposing this for very long jobs?
    minimum_time_to_kill_in_days = 1
    minimum_time_to_kill_in_seconds = minimum_time_to_kill_in_days * 24 * 60 * 60

    # batch_client.create_job_queue() is not synchronous and so we need to wait a bit afterwards
    batch_client.create_job_queue(
        jobQueueName=job_queue_name,
        state="ENABLED",
        priority=1,
        computeEnvironmentOrder=[
            dict(order=1, computeEnvironment=compute_environment_name),
        ],
        # Note: boto3 annotates the reason as a generic string
        # But really it is Literal[
        #    "MISCONFIGURATION:COMPUTE_ENVIRONMENT_MAX_RESOURCE", "MISCONFIGURATION:JOB_RESOURCE_REQUIREMENT"
        # ]
        # And we should have limits on both
        jobStateTimeLimitActions=[
            dict(
                reason="MISCONFIGURATION:COMPUTE_ENVIRONMENT_MAX_RESOURCE",
                state="RUNNABLE",
                maxTimeSeconds=minimum_time_to_kill_in_seconds,
                action="CANCEL",
            ),
            dict(
                reason="MISCONFIGURATION:JOB_RESOURCE_REQUIREMENT",
                state="RUNNABLE",
                maxTimeSeconds=minimum_time_to_kill_in_seconds,
                action="CANCEL",
            ),
        ],
    )

    job_queue_request = batch_client.describe_job_queues(jobQueues=[job_queue_name])
    job_queue_response = job_queue_request["jobQueues"]
    job_queue_status = job_queue_response[0]["status"] if len(job_queue_response) == 1 else ""
    retry_count = 0
    while job_queue_status != "VALID" and retry_count <= max_retries:
        retry_count += 1
        time.sleep(10)
        job_queue_request = batch_client.describe_job_queues(jobQueues=[job_queue_name])
        job_queue_response = job_queue_request["jobQueues"]
        job_queue_status = job_queue_response[0]["status"] if len(job_queue_response) == 1 else ""

    if len(job_queue_response) != 1 and job_queue_status != "VALID":
        raise ValueError(f"Job queue '{job_queue_status}' failed to launch after {max_retries} retries.")

    return None


def _create_or_get_efs_id(
    efs_volume_name: Optional[str], efs_client: "boto3.client.efs", region: str = "us-east-2"
) -> Optional[str]:  # pragma: no cover
    if efs_volume_name is None:
        return None

    if region != "us-east-2":
        raise NotImplementedError("EFS volumes are only supported in us-east-2 for now.")

    available_efs_volumes = efs_client.describe_file_systems()
    matching_efs_volumes = [
        file_system
        for file_system in available_efs_volumes["FileSystems"]
        for tag in file_system["Tags"]
        if tag["Key"] == "Name" and tag["Value"] == efs_volume_name
    ]

    if len(matching_efs_volumes) > 1:
        efs_volume = matching_efs_volumes[0]
        efs_id = efs_volume["FileSystemId"]

        return efs_id

    # Existing volume not found - must create a fresh one and set mount targets on it
    efs_volume = efs_client.create_file_system(
        PerformanceMode="generalPurpose",  # Only type supported in one-zone
        Encrypted=False,
        ThroughputMode="elastic",
        # TODO: figure out how to make job spawn only on subregion for OneZone discount
        # AvailabilityZoneName=subregion,
        Backup=False,
        Tags=[{"Key": "Name", "Value": efs_volume_name}],
    )
    efs_id = efs_volume["FileSystemId"]

    # Takes a while to spin up - cannot assign mount targets until it is ready
    # TODO: in a follow-up replace with more robust checking mechanism
    time.sleep(60)

    # TODO: in follow-up, figure out how to fetch this automatically and from any region
    # (might even resolve those previous OneZone issues)
    region_to_subnet_id = {
        "us-east-2a": "subnet-0890a93aedb42e73e",
        "us-east-2b": "subnet-0e20bbcfb951b5387",
        "us-east-2c": "subnet-0680e07980538b786",
    }
    for subnet_id in region_to_subnet_id.values():
        efs_client.create_mount_target(
            FileSystemId=efs_id,
            SubnetId=subnet_id,
            SecurityGroups=[
                "sg-001699e5b7496b226",
            ],
        )
    time.sleep(60)  # Also takes a while to create the mount targets so add some buffer time

    return efs_id


def _generate_job_definition_name(
    *,
    docker_image: str,
    minimum_worker_ram_in_gib: int,
    minimum_worker_cpus: int,
    efs_id: Optional[str] = None,
) -> str:  # pragma: no cover
    """
    Generate a job definition name for the AWS Batch job.

    Note that Docker images don't strictly require a tag to be pulled or used - 'latest' is always used by default.

    Parameters
    ----------
    docker_image : str
        The name of the Docker image to use for the job.
    minimum_worker_ram_in_gib : int
        The minimum amount of base worker memory required to run this job.
        Determines the EC2 instance type selected by the automatic 'best fit' selector.
        Recommended to be several GiB to allow comfortable buffer space for data chunk iterators.
    minimum_worker_cpus : int
        The minimum number of CPUs required to run this job.
        A minimum of 4 is required, even if only one will be used in the actual process.
    """
    docker_tags = docker_image.split(":")[1:]
    docker_tag = docker_tags[0] if len(docker_tags) > 1 else None
    parsed_docker_image_name = docker_image.replace(":", "-")  # AWS Batch does not allow colons in job definition names

    job_definition_name = f"neuroconv_batch"
    job_definition_name += f"_{parsed_docker_image_name}-image"
    job_definition_name += f"_{minimum_worker_ram_in_gib}-GiB-RAM"
    job_definition_name += f"_{minimum_worker_cpus}-CPU"
    if efs_id is not None:
        job_definition_name += f"_{efs_id}"
    if docker_tag is None or docker_tag == "latest":
        date = datetime.now().strftime("%Y-%m-%d")
        job_definition_name += f"_created-on-{date}"

    return job_definition_name


def _ensure_job_definition_exists_and_get_arn(
    *,
    job_definition_name: str,
    docker_image: str,
    minimum_worker_ram_in_gib: int,
    minimum_worker_cpus: int,
    role_info: dict,
    batch_client: "boto3.client.Batch",
    efs_id: Optional[str] = None,
    max_retries: int = 12,
) -> str:  # pragma: no cover
    """
    Ensure that the job definition exists in AWS Batch and return the official ARN.

    Automatically generates a job definition name using the docker image, its tags, and worker configuration.
    The creation date is also appended if the docker image was not tagged or was tagged as 'latest'.

    Note that registering a job definition does not require either a compute environment or a job queue to exist.

    Also note that registration is strict; once created, a job definition is tagged with ':< revision index >' and
    any subsequent changes create a new revision index (and technically a new definition as well).

    Parameters
    ----------
    job_definition_name : str
        The name of the job definition to use for the job.
    docker_image : str
        The name of the Docker image to use for the job.
    minimum_worker_ram_in_gib : int
        The minimum amount of base worker memory required to run this job.
        Determines the EC2 instance type selected by the automatic 'best fit' selector.
        Recommended to be several GiB to allow comfortable buffer space for data chunk iterators.
    minimum_worker_cpus : int
        The minimum number of CPUs required to run this job.
        A minimum of 4 is required, even if only one will be used in the actual process.
    role_info : dict
        The IAM role information for the job.
    batch_client : boto3.client.Batch
        The AWS Batch client to use for the job.
    efs_id : str, optional
        The EFS volume information for the job.
        The path exposed to the container will always be `/mnt/efs`.
    max_retries : int, default: 12
        If the job definition does not already exist, then this is the maximum number of times to synchronously
        check for its successful creation before erroring.
        This is essential for a clean setup of the entire pipeline, or else later steps might error because they tried
        to launch before the job definition was ready.

    Returns
    -------
    job_definition_arn : str
        The full ARN of the job definition.
    """
    revision = 1
    job_definition_with_revision = f"{job_definition_name}:{revision}"
    job_definition_request = batch_client.describe_job_definitions(jobDefinitions=[job_definition_with_revision])
    job_definition_response = job_definition_request["jobDefinitions"]

    # Increment revision until we either find one that is active or we fail to find one that exists
    while len(job_definition_response) == 1:
        if job_definition_response[0]["status"] == "ACTIVE":
            return job_definition_response[0]["jobDefinitionArn"]
        else:
            revision += 1

            job_definition_with_revision = f"{job_definition_name}:{revision}"
            job_definition_request = batch_client.describe_job_definitions(
                jobDefinitions=[job_definition_with_revision]
            )
            job_definition_response = job_definition_request["jobDefinitions"]

    resource_requirements = [
        {
            "value": str(int(minimum_worker_ram_in_gib * 1024)),  # boto3 expects memory in round MiB
            "type": "MEMORY",
        },
        {"value": str(minimum_worker_cpus), "type": "VCPU"},
    ]

    minimum_time_to_kill_in_days = 1  # Note: eventually consider exposing this for very long jobs?
    minimum_time_to_kill_in_seconds = minimum_time_to_kill_in_days * 24 * 60 * 60

    volumes = []
    mountPoints = []
    if efs_id is not None:
        volumes = [
            {
                "name": "neuroconv_batch_efs_mounted",
                "efsVolumeConfiguration": {
                    "fileSystemId": efs_id,
                    "transitEncryption": "DISABLED",
                },
            },
        ]
        mountPoints = [{"containerPath": "/mnt/efs/", "readOnly": False, "sourceVolume": "neuroconv_batch_efs_mounted"}]

    # batch_client.register_job_definition() is not synchronous and so we need to wait a bit afterwards
    batch_client.register_job_definition(
        jobDefinitionName=job_definition_name,
        type="container",
        timeout=dict(attemptDurationSeconds=minimum_time_to_kill_in_seconds),
        containerProperties=dict(
            image=docker_image,
            resourceRequirements=resource_requirements,
            # TODO: investigate if any IAM role is explicitly needed in conjunction with the credentials
            # jobRoleArn=role_info["Role"]["Arn"],
            # executionRoleArn=role_info["Role"]["Arn"],
            volumes=volumes,
            mountPoints=mountPoints,
        ),
        platformCapabilities=["EC2"],
    )

    job_definition_request = batch_client.describe_job_definitions(jobDefinitions=[job_definition_with_revision])
    job_definition_response = job_definition_request["jobDefinitions"]
    job_definition_status = job_definition_response[0]["status"] if len(job_definition_response) == 1 else ""
    retry_count = 0
    while job_definition_status != "ACTIVE" and retry_count <= max_retries:
        retry_count += 1
        time.sleep(10)
        job_definition_request = batch_client.describe_job_definitions(jobDefinitions=[job_definition_with_revision])
        job_definition_response = job_definition_request["jobDefinitions"]
        job_definition_status = job_definition_response[0]["status"] if len(job_definition_response) == 1 else ""

    if len(job_definition_response) != 1 or job_definition_status != "ACTIVE":
        raise ValueError(
            f"Job definition '{job_definition_with_revision}' failed to launch after {max_retries} retries."
        )

    return job_definition_response[0]["jobDefinitionArn"]



from copy import deepcopy
from typing import Optional

from pydantic import FilePath, validate_call
from pynwb import NWBFile

from neuroconv import NWBConverter
from neuroconv.datainterfaces import LightningPoseDataInterface, VideoInterface
from neuroconv.tools.nwb_helpers import make_or_load_nwbfile
from neuroconv.utils import (
    DeepDict,
    dict_deep_update,
    get_schema_from_method_signature,
)


class LightningPoseConverter(NWBConverter):
    """Primary conversion class for handling Lightning Pose data streams."""

    display_name = "Lightning Pose Converter"
    keywords = ("pose estimation", "video")
    associated_suffixes = (".csv", ".mp4")
    info = "Interface for handling multiple streams of lightning pose data."

    @classmethod
    def get_source_schema(cls):
        return get_schema_from_method_signature(cls)

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        original_video_file_path: FilePath,
        labeled_video_file_path: Optional[FilePath] = None,
        image_series_original_video_name: Optional[str] = None,
        image_series_labeled_video_name: Optional[str] = None,
        verbose: bool = True,
    ):
        """
        The converter for Lightning Pose format to convert the pose estimation data
        along with the original and the optional labeled video added as ImageSeries to NWB.

        Parameters
        ----------
        file_path : FilePath
            Path to the .csv file that contains the predictions from Lightning Pose.
        original_video_file_path : FilePath
            Path to the original video file (.mp4).
        labeled_video_file_path : FilePath, optional
            Path to the labeled video file (.mp4).
        image_series_original_video_name: string, optional
            The name of the ImageSeries to add for the original video.
        image_series_labeled_video_name: string, optional
            The name of the ImageSeries to add for the labeled video.
        verbose : bool, default: True
            controls verbosity. ``True`` by default.
        """
        self.verbose = verbose
        self.data_interface_objects = dict(
            OriginalVideo=VideoInterface(file_paths=[original_video_file_path]),
            PoseEstimation=LightningPoseDataInterface(
                file_path=file_path,
                original_video_file_path=original_video_file_path,
                labeled_video_file_path=labeled_video_file_path,
            ),
        )
        self.original_video_name = image_series_original_video_name or "ImageSeriesOriginalVideo"
        self.labeled_video_name = None
        if labeled_video_file_path:
            self.labeled_video_name = image_series_labeled_video_name or "ImageSeriesLabeledVideo"
            self.data_interface_objects.update(dict(LabeledVideo=VideoInterface(file_paths=[labeled_video_file_path])))

    def get_conversion_options_schema(self) -> dict:
        conversion_options_schema = get_schema_from_method_signature(
            method=self.add_to_nwbfile, exclude=["nwbfile", "metadata"]
        )

        return conversion_options_schema

    def get_metadata(self) -> DeepDict:
        metadata = self.data_interface_objects["PoseEstimation"].get_metadata()
        original_video_interface = self.data_interface_objects["OriginalVideo"]
        original_videos_metadata = original_video_interface.get_metadata()
        metadata = dict_deep_update(metadata, original_videos_metadata)

        original_videos_metadata["Behavior"]["Videos"][0].update(
            name=self.original_video_name,
            description="The original video used for pose estimation.",
        )

        if "LabeledVideo" in self.data_interface_objects:
            labeled_video_interface = self.data_interface_objects["LabeledVideo"]
            labeled_videos_metadata = labeled_video_interface.get_metadata()
            labeled_videos_metadata["Behavior"]["Videos"][0].update(
                name=self.labeled_video_name,
                description="The video recorded by camera with the pose estimation labels.",
            )

            metadata = dict_deep_update(metadata, labeled_videos_metadata)

        return metadata

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: dict,
        reference_frame: Optional[str] = None,
        confidence_definition: Optional[str] = None,
        external_mode: bool = True,
        starting_frames_original_videos: Optional[list[int]] = None,
        starting_frames_labeled_videos: Optional[list[int]] = None,
        stub_test: bool = False,
    ):
        original_video_interface = self.data_interface_objects["OriginalVideo"]

        original_video_metadata = next(
            video_metadata
            for video_metadata in metadata["Behavior"]["Videos"]
            if video_metadata["name"] == self.original_video_name
        )
        if original_video_metadata is None:
            raise ValueError(f"Metadata for '{self.original_video_name}' not found in metadata['Behavior']['Videos'].")
        metadata_copy = deepcopy(metadata)
        metadata_copy["Behavior"]["Videos"] = [original_video_metadata]
        original_video_interface.add_to_nwbfile(
            nwbfile=nwbfile,
            metadata=metadata_copy,
            stub_test=stub_test,
            external_mode=external_mode,
            starting_frames=starting_frames_original_videos,
        )

        if "LabeledVideo" in self.data_interface_objects:
            labeled_video_interface = self.data_interface_objects["LabeledVideo"]
            labeled_video_metadata = next(
                video_metadata
                for video_metadata in metadata["Behavior"]["Videos"]
                if video_metadata["name"] == self.labeled_video_name
            )
            if labeled_video_metadata is None:
                raise ValueError(
                    f"Metadata for '{self.labeled_video_name}' not found in metadata['Behavior']['Videos']."
                )
            metadata_copy["Behavior"]["Videos"] = [labeled_video_metadata]
            labeled_video_interface.add_to_nwbfile(
                nwbfile=nwbfile,
                metadata=metadata_copy,
                stub_test=stub_test,
                external_mode=external_mode,
                starting_frames=starting_frames_labeled_videos,
                module_name="behavior",
            )

        self.data_interface_objects["PoseEstimation"].add_to_nwbfile(
            nwbfile=nwbfile,
            metadata=metadata,
            reference_frame=reference_frame,
            confidence_definition=confidence_definition,
            stub_test=stub_test,
        )

    def run_conversion(
        self,
        nwbfile_path: Optional[FilePath] = None,
        nwbfile: Optional[NWBFile] = None,
        metadata: Optional[dict] = None,
        overwrite: bool = False,
        reference_frame: Optional[str] = None,
        confidence_definition: Optional[str] = None,
        external_mode: bool = True,
        starting_frames_original_videos: Optional[list] = None,
        starting_frames_labeled_videos: Optional[list] = None,
        stub_test: bool = False,
    ) -> None:
        if metadata is None:
            metadata = self.get_metadata()

        self.validate_metadata(metadata=metadata)

        self.temporally_align_data_interfaces()

        with make_or_load_nwbfile(
            nwbfile_path=nwbfile_path,
            nwbfile=nwbfile,
            metadata=metadata,
            overwrite=overwrite,
            verbose=self.verbose,
        ) as nwbfile_out:
            self.add_to_nwbfile(
                nwbfile=nwbfile_out,
                metadata=metadata,
                reference_frame=reference_frame,
                confidence_definition=confidence_definition,
                external_mode=external_mode,
                starting_frames_original_videos=starting_frames_original_videos,
                starting_frames_labeled_videos=starting_frames_labeled_videos,
                stub_test=stub_test,
            )



from .roiextractors import (
    _check_if_imaging_fits_into_memory,
    get_nwb_imaging_metadata,
    get_nwb_segmentation_metadata,
    add_background_fluorescence_traces,
    add_background_plane_segmentation,
    add_devices_to_nwbfile,
    add_fluorescence_traces_to_nwbfile,
    add_image_segmentation_to_nwbfile,
    add_imaging_to_nwbfile,
    add_imaging_plane_to_nwbfile,
    add_photon_series_to_nwbfile,
    add_plane_segmentation_to_nwbfile,
    add_segmentation_to_nwbfile,
    add_summary_images_to_nwbfile,
    write_imaging_to_nwbfile,
    write_segmentation_to_nwbfile,
    add_devices,
    add_fluorescence_traces,
    add_image_segmentation,
    add_imaging,
    add_imaging_plane,
    add_photon_series,
    add_plane_segmentation,
    add_segmentation,
    add_summary_images,
    write_imaging,
    write_segmentation,
)



"""Contains core class definitions for the NWBConverter and ConverterPipe."""

import json
import warnings
from collections import Counter
from pathlib import Path
from typing import Literal, Optional, Union

from jsonschema import validate
from pydantic import FilePath, validate_call
from pynwb import NWBFile

from .basedatainterface import BaseDataInterface
from .tools.nwb_helpers import (
    HDF5BackendConfiguration,
    ZarrBackendConfiguration,
    configure_backend,
    get_default_backend_configuration,
    get_default_nwbfile_metadata,
    make_nwbfile_from_metadata,
    make_or_load_nwbfile,
)
from .tools.nwb_helpers._metadata_and_file_helpers import _resolve_backend
from .utils import (
    dict_deep_update,
    fill_defaults,
    get_base_schema,
    load_dict_from_file,
    unroot_schema,
)
from .utils.dict import DeepDict
from .utils.json_schema import _NWBMetaDataEncoder, _NWBSourceDataEncoder


class NWBConverter:
    """Primary class for all NWB conversion classes."""

    display_name: Union[str, None] = None
    keywords: tuple[str] = tuple()
    associated_suffixes: tuple[str] = tuple()
    info: Union[str, None] = None

    data_interface_classes = None

    @classmethod
    def get_source_schema(cls) -> dict:
        """Compile input schemas from each of the data interface classes."""
        source_schema = get_base_schema(
            root=True,
            id_="source.schema.json",
            title="Source data schema",
            description="Schema for the source data, files and directories",
            version="0.1.0",
        )
        for interface_name, data_interface in cls.data_interface_classes.items():
            source_schema["properties"].update({interface_name: unroot_schema(data_interface.get_source_schema())})
        return source_schema

    @classmethod
    def validate_source(cls, source_data: dict[str, dict], verbose: bool = True):
        """Validate source_data against Converter source_schema."""
        cls._validate_source_data(source_data=source_data, verbose=verbose)

    def _validate_source_data(self, source_data: dict[str, dict], verbose: bool = True):

        encoder = _NWBSourceDataEncoder()
        # The encoder produces a serialized object, so we deserialized it for comparison

        serialized_source_data = encoder.encode(source_data)
        decoded_source_data = json.loads(serialized_source_data)

        validate(instance=decoded_source_data, schema=self.get_source_schema())
        if verbose:
            print("Source data is valid!")

    @validate_call
    def __init__(self, source_data: dict[str, dict], verbose: bool = True):
        """Validate source_data against source_schema and initialize all data interfaces."""
        self.verbose = verbose
        self._validate_source_data(source_data=source_data, verbose=self.verbose)
        self.data_interface_objects = {
            name: data_interface(**source_data[name])
            for name, data_interface in self.data_interface_classes.items()
            if name in source_data
        }

    def get_metadata_schema(self) -> dict:
        """Compile metadata schemas from each of the data interface objects."""
        metadata_schema = load_dict_from_file(Path(__file__).parent / "schemas" / "base_metadata_schema.json")
        for data_interface in self.data_interface_objects.values():
            interface_schema = unroot_schema(data_interface.get_metadata_schema())
            metadata_schema = dict_deep_update(metadata_schema, interface_schema)

        default_values = self.get_metadata()
        fill_defaults(metadata_schema, default_values)
        return metadata_schema

    def get_metadata(self) -> DeepDict:
        """Auto-fill as much of the metadata as possible. Must comply with metadata schema."""
        metadata = get_default_nwbfile_metadata()
        for interface in self.data_interface_objects.values():
            interface_metadata = interface.get_metadata()
            metadata = dict_deep_update(metadata, interface_metadata)
        return metadata

    def validate_metadata(self, metadata: dict[str, dict], append_mode: bool = False):
        """Validate metadata against Converter metadata_schema."""
        encoder = _NWBMetaDataEncoder()
        # The encoder produces a serialized object, so we deserialized it for comparison
        serialized_metadata = encoder.encode(metadata)
        decoded_metadata = json.loads(serialized_metadata)

        metadata_schema = self.get_metadata_schema()
        if append_mode:
            # Eliminate required from NWBFile
            nwbfile_schema = metadata_schema["properties"]["NWBFile"]
            nwbfile_schema.pop("required", None)

        validate(instance=decoded_metadata, schema=metadata_schema)
        if self.verbose:
            print("Metadata is valid!")

    def get_conversion_options_schema(self) -> dict:
        """Compile conversion option schemas from each of the data interface classes."""
        conversion_options_schema = get_base_schema(
            root=True,
            id_="conversion_options.schema.json",
            title="Conversion options schema",
            description="Schema for the conversion options",
            version="0.1.0",
        )
        for interface_name, data_interface in self.data_interface_objects.items():
            conversion_options_schema["properties"].update(
                {interface_name: unroot_schema(data_interface.get_conversion_options_schema())}
            )

        return conversion_options_schema

    def validate_conversion_options(self, conversion_options: dict[str, dict]):
        """Validate conversion_options against Converter conversion_options_schema."""
        validate(instance=conversion_options or {}, schema=self.get_conversion_options_schema())
        if self.verbose:
            print("conversion_options is valid!")

    def create_nwbfile(self, metadata: Optional[dict] = None, conversion_options: Optional[dict] = None) -> NWBFile:
        """
        Create and return an in-memory pynwb.NWBFile object with this interface's data added to it.

        Parameters
        ----------
        metadata : dict, optional
            Metadata dictionary with information used to create the NWBFile.
        conversion_options : dict, optional
            Similar to source_data, a dictionary containing keywords for each interface for which non-default
            conversion specification is requested.

        Returns
        -------
        nwbfile : pynwb.NWBFile
            The in-memory object with this interface's data added to it.
        """
        if metadata is None:
            metadata = self.get_metadata()

        nwbfile = make_nwbfile_from_metadata(metadata=metadata)
        self.add_to_nwbfile(nwbfile=nwbfile, metadata=metadata, conversion_options=conversion_options)
        return nwbfile

    def add_to_nwbfile(self, nwbfile: NWBFile, metadata, conversion_options: Optional[dict] = None) -> None:
        conversion_options = conversion_options or dict()
        for interface_name, data_interface in self.data_interface_objects.items():
            data_interface.add_to_nwbfile(
                nwbfile=nwbfile, metadata=metadata, **conversion_options.get(interface_name, dict())
            )

    def run_conversion(
        self,
        nwbfile_path: Optional[FilePath] = None,
        nwbfile: Optional[NWBFile] = None,
        metadata: Optional[dict] = None,
        overwrite: bool = False,
        # TODO: when all H5DataIO prewraps are gone, introduce Zarr safely
        # backend: Union[Literal["hdf5", "zarr"]],
        # backend_configuration: Optional[Union[HDF5BackendConfiguration, ZarrBackendConfiguration]] = None,
        backend: Optional[Literal["hdf5"]] = None,
        backend_configuration: Optional[HDF5BackendConfiguration] = None,
        conversion_options: Optional[dict] = None,
    ) -> None:
        """
        Run the NWB conversion over all the instantiated data interfaces.

        Parameters
        ----------
        nwbfile_path : FilePathType
            Path for where to write or load (if overwrite=False) the NWBFile.
            If specified, the context will always write to this location.
        nwbfile : NWBFile, optional
            An in-memory NWBFile object to write to the location.
        metadata : dict, optional
            Metadata dictionary with information used to create the NWBFile when one does not exist or overwrite=True.
        overwrite : bool, default: False
            Whether to overwrite the NWBFile if one exists at the nwbfile_path.
            The default is False (append mode).
        backend : "hdf5", optional
            The type of backend to use when writing the file.
            If a `backend_configuration` is not specified, the default type will be "hdf5".
            If a `backend_configuration` is specified, then the type will be auto-detected.
        backend_configuration : HDF5BackendConfiguration, optional
            The configuration model to use when configuring the datasets for this backend.
            To customize, call the `.get_default_backend_configuration(...)` method, modify the returned
            BackendConfiguration object, and pass that instead.
            Otherwise, all datasets will use default configuration settings.
        conversion_options : dict, optional
            Similar to source_data, a dictionary containing keywords for each interface for which non-default
            conversion specification is requested.
        """

        if nwbfile_path is None:
            warnings.warn(  # TODO: remove on or after 2024/12/26
                "Using Converter.run_conversion without specifying nwbfile_path is deprecated. To create an "
                "NWBFile object in memory, use Converter.create_nwbfile. To append to an existing NWBFile object,"
                " use Converter.add_to_nwbfile."
            )

        backend = _resolve_backend(backend, backend_configuration)
        no_nwbfile_provided = nwbfile is None  # Otherwise, variable reference may mutate later on inside the context

        file_initially_exists = Path(nwbfile_path).exists() if nwbfile_path is not None else False
        append_mode = file_initially_exists and not overwrite

        if metadata is None:
            metadata = self.get_metadata()

        self.validate_metadata(metadata=metadata, append_mode=append_mode)
        self.validate_conversion_options(conversion_options=conversion_options)

        self.temporally_align_data_interfaces()

        with make_or_load_nwbfile(
            nwbfile_path=nwbfile_path,
            nwbfile=nwbfile,
            metadata=metadata,
            overwrite=overwrite,
            backend=backend,
            verbose=getattr(self, "verbose", False),
        ) as nwbfile_out:
            if no_nwbfile_provided:
                self.add_to_nwbfile(nwbfile=nwbfile_out, metadata=metadata, conversion_options=conversion_options)

            if backend_configuration is None:
                backend_configuration = self.get_default_backend_configuration(nwbfile=nwbfile_out, backend=backend)

            configure_backend(nwbfile=nwbfile_out, backend_configuration=backend_configuration)

    def temporally_align_data_interfaces(self):
        """Override this method to implement custom alignment."""
        pass

    @staticmethod
    def get_default_backend_configuration(
        nwbfile: NWBFile,
        backend: Literal["hdf5", "zarr"] = "hdf5",
    ) -> Union[HDF5BackendConfiguration, ZarrBackendConfiguration]:
        """
        Fill and return a default backend configuration to serve as a starting point for further customization.

        Parameters
        ----------
        nwbfile : pynwb.NWBFile
            The in-memory object with this interface's data already added to it.
        backend : "hdf5" or "zarr", default: "hdf5"
            The type of backend to use when creating the file.

        Returns
        -------
        backend_configuration : HDF5BackendConfiguration or ZarrBackendConfiguration
            The default configuration for the specified backend type.
        """
        return get_default_backend_configuration(nwbfile=nwbfile, backend=backend)


class ConverterPipe(NWBConverter):
    """Takes a list or dict of pre-initialized interfaces as arguments to build an NWBConverter class."""

    @classmethod
    def get_source_schema(cls) -> dict:
        raise NotImplementedError("Source data not available with previously initialized classes.")

    @classmethod
    def validate_source(cls):
        raise NotImplementedError("Source data not available with previously initialized classes.")

    def __init__(self, data_interfaces: Union[list[BaseDataInterface], dict[str, BaseDataInterface]], verbose=True):
        self.verbose = verbose
        if isinstance(data_interfaces, list):
            # Create unique names for each interface
            counter = {interface.__class__.__name__: 0 for interface in data_interfaces}
            total_counts = Counter([interface.__class__.__name__ for interface in data_interfaces])
            self.data_interface_objects = dict()
            for interface in data_interfaces:
                class_name = interface.__class__.__name__
                counter[class_name] += 1
                unique_signature = f"{counter[class_name]:03}" if total_counts[class_name] > 1 else ""
                interface_name = f"{class_name}{unique_signature}"
                self.data_interface_objects[interface_name] = interface
        elif isinstance(data_interfaces, dict):
            self.data_interface_objects = data_interfaces

        self.data_interface_classes = {
            name: interface.__class__ for name, interface in self.data_interface_objects.items()
        }

    def get_conversion_options_schema(self) -> dict:
        """Compile conversion option schemas from each of the data interface classes."""
        conversion_options_schema = get_base_schema(
            root=True,
            id_="conversion_options.schema.json",
            title="Conversion options schema",
            description="Schema for the conversion options",
            version="0.1.0",
        )
        for interface_name, data_interface in self.data_interface_objects.items():

            schema = data_interface.get_conversion_options_schema()
            conversion_options_schema["properties"].update({interface_name: unroot_schema(schema)})
        return conversion_options_schema



import json
from pathlib import Path
from typing import Literal, Optional

import numpy as np
import scipy
from pydantic import FilePath, validate_call
from pynwb import NWBFile

from ....basetemporalalignmentinterface import BaseTemporalAlignmentInterface
from ....tools.audio import add_acoustic_waveform_series
from ....utils import (
    get_base_schema,
)


def _check_audio_names_are_unique(metadata: dict):
    neurodata_names = [neurodata["name"] for neurodata in metadata]
    neurodata_names_are_unique = len(set(neurodata_names)) == len(neurodata_names)
    assert neurodata_names_are_unique, "Some of the names for Audio metadata are not unique."


class AudioInterface(BaseTemporalAlignmentInterface):
    """Data interface for writing .wav audio recordings to an NWB file."""

    display_name = "Wav Audio"
    keywords = ("sound", "microphone")
    associated_suffixes = (".wav",)
    info = "Interface for writing audio recordings to an NWB file."

    @validate_call
    def __init__(self, file_paths: list[FilePath], verbose: bool = False):
        """
        Data interface for writing acoustic recordings to an NWB file.

        Writes acoustic recordings as an ``AcousticWaveformSeries`` from the ndx_sound extension.

        Parameters
        ----------
        file_paths : list of FilePathTypes
            The file paths to the audio recordings in sorted, consecutive order.
            We recommend using ``natsort`` to ensure the files are in consecutive order::

                from natsort import natsorted
                natsorted(file_paths)

        verbose : bool, default: False
        """
        suffixes = [suffix for file_path in file_paths for suffix in Path(file_path).suffixes]
        format_is_not_supported = [
            suffix for suffix in suffixes if suffix not in [".wav"]
        ]  # TODO: add support for more formats
        if format_is_not_supported:
            raise ValueError(
                "The currently supported file format for audio is WAV file. "
                f"Some of the provided files does not match this format: {format_is_not_supported}."
            )

        self._number_of_audio_files = len(file_paths)
        self.verbose = verbose
        super().__init__(file_paths=file_paths)
        self._segment_starting_times = None

    def get_metadata_schema(self) -> dict:
        metadata_schema = super().get_metadata_schema()

        time_series_metadata_schema_path = (
            Path(__file__).parent.parent.parent.parent / "schemas" / "time_series_schema.json"
        )
        with open(file=time_series_metadata_schema_path) as fp:
            time_series_metadata_schema = json.load(fp=fp)
        time_series_metadata_schema.update(required=["name"])

        metadata_schema["properties"]["Behavior"] = get_base_schema(tag="Behavior")
        metadata_schema["properties"]["Behavior"].update(
            required=["Audio"],
            properties=dict(
                Audio=dict(
                    type="array",
                    minItems=1,
                    items=time_series_metadata_schema,
                )
            ),
        )
        return metadata_schema

    def get_metadata(self) -> dict:
        default_name = "AcousticWaveformSeries"
        is_multiple_file_path = len(self.source_data["file_paths"]) > 1
        audio_metadata = [
            dict(
                name=default_name + str(file_ind) if is_multiple_file_path and file_ind > 0 else default_name,
                description="Acoustic waveform series.",
            )
            for file_ind, file_path in enumerate(self.source_data["file_paths"])
        ]
        behavior_metadata = dict(Audio=audio_metadata)

        metadata = super().get_metadata()
        metadata.update(Behavior=behavior_metadata)
        return metadata

    def get_original_timestamps(self) -> np.ndarray:
        raise NotImplementedError("The AudioInterface does not yet support timestamps.")

    def get_timestamps(self) -> Optional[np.ndarray]:
        raise NotImplementedError("The AudioInterface does not yet support timestamps.")

    def set_aligned_timestamps(self, aligned_timestamps: list[np.ndarray]):
        raise NotImplementedError("The AudioInterface does not yet support timestamps.")

    def set_aligned_starting_time(self, aligned_starting_time: float):
        """
        Align all starting times for all audio files in this interface relative to the common session start time.

        Must be in units seconds relative to the common 'session_start_time'.

        Parameters
        ----------
        aligned_starting_time : float
            The common starting time for all temporal data in this interface.
            Applies to all segments if there are multiple file paths used by the interface.
        """
        if self._segment_starting_times is None and self._number_of_audio_files == 1:
            self._segment_starting_times = [aligned_starting_time]
        elif self._segment_starting_times is not None and self._number_of_audio_files > 1:
            self._segment_starting_times = [
                segment_starting_time + aligned_starting_time for segment_starting_time in self._segment_starting_times
            ]
        else:
            raise ValueError(
                "There are no segment starting times to shift by a common value! "
                "Please set them using 'set_aligned_segment_starting_times'."
            )

    def set_aligned_segment_starting_times(self, aligned_segment_starting_times: list[float]):
        """
        Align the individual starting time for each audio file in this interface relative to the common session start time.

        Must be in units seconds relative to the common 'session_start_time'.

        Parameters
        ----------
        aligned_segment_starting_times : list of floats
            The relative starting times of each audio file (segment).
        """
        aligned_segment_starting_times_length = len(aligned_segment_starting_times)
        assert isinstance(aligned_segment_starting_times, list) and all(
            [isinstance(x, float) for x in aligned_segment_starting_times]
        ), "Argument 'aligned_segment_starting_times' must be a list of floats."
        assert aligned_segment_starting_times_length == self._number_of_audio_files, (
            f"The number of entries in 'aligned_segment_starting_times' ({aligned_segment_starting_times_length}) "
            f"must be equal to the number of audio file paths ({self._number_of_audio_files})."
        )

        self._segment_starting_times = aligned_segment_starting_times

    def align_by_interpolation(self, unaligned_timestamps: np.ndarray, aligned_timestamps: np.ndarray):
        raise NotImplementedError("The AudioInterface does not yet support timestamps.")

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: Optional[dict] = None,
        stub_test: bool = False,
        stub_frames: int = 1000,
        write_as: Literal["stimulus", "acquisition"] = "stimulus",
        iterator_options: Optional[dict] = None,
        compression_options: Optional[dict] = None,  # TODO: remove completely after 10/1/2024
        overwrite: bool = False,
        verbose: bool = True,
    ):
        """
        Parameters
        ----------
        nwbfile : NWBFile
            Append to this NWBFile object
        metadata : dict, optional
        stub_test : bool, default: False
        stub_frames : int, default: 1000
        write_as : {'stimulus', 'acquisition'}
            The acoustic waveform series can be added to the NWB file either as
            "stimulus" or as "acquisition".
        iterator_options : dict, optional
            Dictionary of options for the SliceableDataChunkIterator.
        overwrite : bool, default: False
        verbose : bool, default: True

        Returns
        -------
        NWBFile
        """
        file_paths = self.source_data["file_paths"]
        audio_metadata = metadata["Behavior"]["Audio"]
        _check_audio_names_are_unique(metadata=audio_metadata)
        assert len(audio_metadata) == self._number_of_audio_files, (
            f"The Audio metadata is incomplete ({len(audio_metadata)} entry)! "
            f"Expected {self._number_of_audio_files} (one for each entry of 'file_paths')."
        )

        audio_name_list = [audio["name"] for audio in audio_metadata]
        any_duplicated_audio_names = len(set(audio_name_list)) < len(file_paths)
        if any_duplicated_audio_names:
            raise ValueError("There are duplicated file names in the metadata!")

        if self._number_of_audio_files > 1 and self._segment_starting_times is None:
            raise ValueError(
                "If you have multiple audio files, then you must specify each starting time by calling "
                "'.set_aligned_segment_starting_times(...)'!"
            )
        starting_times = self._segment_starting_times or [0.0]

        for file_index, (acoustic_waveform_series_metadata, file_path) in enumerate(zip(audio_metadata, file_paths)):
            sampling_rate, acoustic_series = scipy.io.wavfile.read(filename=file_path, mmap=True)

            if stub_test:
                acoustic_series = acoustic_series[:stub_frames]

            add_acoustic_waveform_series(
                acoustic_series=acoustic_series,
                nwbfile=nwbfile,
                rate=sampling_rate,
                metadata=acoustic_waveform_series_metadata,
                write_as=write_as,
                starting_time=starting_times[file_index],
                iterator_options=iterator_options,
                compression_options=compression_options,  # TODO: remove completely after 10/1/2024; still passing for deprecation warning
            )

        return nwbfile



"""Collection of Pydantic models and helper functions for configuring dataset IO parameters for different backends."""

# Mark these imports as private to avoid polluting the namespace; only used in global BACKEND_NWB_IO mapping


from ._backend_configuration import (
    BACKEND_CONFIGURATIONS,
    get_default_backend_configuration,
)
from ._configuration_models import DATASET_IO_CONFIGURATIONS
from ._configuration_models._base_backend import BackendConfiguration
from ._configuration_models._base_dataset_io import DatasetIOConfiguration
from ._configuration_models._hdf5_backend import HDF5BackendConfiguration
from ._configuration_models._hdf5_dataset_io import (
    AVAILABLE_HDF5_COMPRESSION_METHODS,
    HDF5DatasetIOConfiguration,
)
from ._configuration_models._zarr_backend import ZarrBackendConfiguration
from ._configuration_models._zarr_dataset_io import (
    AVAILABLE_ZARR_COMPRESSION_METHODS,
    ZarrDatasetIOConfiguration,
)
from ._configure_backend import configure_backend
from ._dataset_configuration import get_default_dataset_io_configurations
from ._metadata_and_file_helpers import (
    BACKEND_NWB_IO,
    add_device_from_metadata,
    configure_and_write_nwbfile,
    get_default_nwbfile_metadata,
    get_module,
    make_nwbfile_from_metadata,
    make_or_load_nwbfile,
)

__all__ = [
    "AVAILABLE_HDF5_COMPRESSION_METHODS",
    "AVAILABLE_ZARR_COMPRESSION_METHODS",
    "BACKEND_CONFIGURATIONS",
    "DATASET_IO_CONFIGURATIONS",
    "BACKEND_NWB_IO",
    "BackendConfiguration",
    "HDF5BackendConfiguration",
    "ZarrBackendConfiguration",
    "DatasetIOConfiguration",
    "HDF5DatasetIOConfiguration",
    "ZarrDatasetIOConfiguration",
    "get_default_backend_configuration",
    "get_default_dataset_io_configurations",
    "configure_backend",
    "get_default_dataset_io_configurations",
    "get_default_backend_configuration",
    "add_device_from_metadata",
    "configure_and_write_nwbfile",
    "get_default_nwbfile_metadata",
    "get_module",
    "make_nwbfile_from_metadata",
    "make_or_load_nwbfile",
]



"""Collection of modifications of HDMF functions that are to be tested/used on this repo until propagation upstream."""

import math
import warnings

import numpy as np
from hdmf.data_utils import GenericDataChunkIterator as HDMFGenericDataChunkIterator


class GenericDataChunkIterator(HDMFGenericDataChunkIterator):  # noqa: D101
    # TODO Should this be added to the API?

    def _get_default_buffer_shape(self, buffer_gb: float = 1.0) -> tuple[int]:
        return self.estimate_default_buffer_shape(
            buffer_gb=buffer_gb, chunk_shape=self.chunk_shape, maxshape=self.maxshape, dtype=self.dtype
        )

    # TODO: move this to the core iterator in HDMF so it can be easily swapped out as well as run on its own
    @staticmethod
    def estimate_default_chunk_shape(chunk_mb: float, maxshape: tuple[int, ...], dtype: np.dtype) -> tuple[int, ...]:
        """
        Select chunk shape with size in MB less than the threshold of chunk_mb.

        Keeps the dimensional ratios of the original data.
        """
        # Elevate any overflow warnings to trigger error.
        # This is usually an indicator of something going terribly wrong with the estimation calculations and should be
        # avoided at all costs.
        warnings.filterwarnings(action="error", message="overflow encountered *")

        assert chunk_mb > 0.0, f"chunk_mb ({chunk_mb}) must be greater than zero!"
        # Eventually, Pydantic validation can handle this validation for us

        n_dims = len(maxshape)
        itemsize = dtype.itemsize
        chunk_bytes = chunk_mb * 1e6

        min_maxshape = min(maxshape)
        v = tuple(math.floor(maxshape_axis / min_maxshape) for maxshape_axis in maxshape)
        prod_v = math.prod(v)
        while prod_v * itemsize > chunk_bytes and prod_v != 1:
            non_unit_min_v = min(x for x in v if x != 1)
            v = tuple(math.floor(x / non_unit_min_v) if x != 1 else x for x in v)
            prod_v = math.prod(v)
        k = math.floor((chunk_bytes / (prod_v * itemsize)) ** (1 / n_dims))
        return tuple([min(k * x, maxshape[dim]) for dim, x in enumerate(v)])

    # TODO: move this to the core iterator in HDMF so it can be easily swapped out as well as run on its own
    @staticmethod
    def estimate_default_buffer_shape(
        buffer_gb: float, chunk_shape: tuple[int, ...], maxshape: tuple[int, ...], dtype: np.dtype
    ) -> tuple[int, ...]:
        # Elevate any overflow warnings to trigger error.
        # This is usually an indicator of something going terribly wrong with the estimation calculations and should be
        # avoided at all costs.
        warnings.filterwarnings(action="error", message="overflow encountered *")

        num_axes = len(maxshape)
        chunk_bytes = math.prod(chunk_shape) * dtype.itemsize

        assert num_axes > 0, f"The number of axes ({num_axes}) is less than one!"
        assert buffer_gb > 0, f"buffer_gb ({buffer_gb}) must be greater than zero!"
        assert (
            buffer_gb >= chunk_bytes / 1e9
        ), f"buffer_gb ({buffer_gb}) must be greater than the chunk size ({chunk_bytes / 1e9})!"
        assert all(np.array(chunk_shape) > 0), f"Some dimensions of chunk_shape ({chunk_shape}) are less than zero!"

        # Early termination condition
        if math.prod(maxshape) * dtype.itemsize / 1e9 < buffer_gb:
            return tuple(maxshape)

        # Note: must occur after taking `math.prod` of it in line above; otherwise it triggers an overflow warning
        maxshape = np.array(maxshape)

        buffer_bytes = chunk_bytes
        axis_sizes_bytes = maxshape * dtype.itemsize
        target_buffer_bytes = buffer_gb * 1e9

        # Recording indices of shortest axes for later use
        if num_axes > 1:  # Only store two shortest if more than 1
            smallest_chunk_axis, second_smallest_chunk_axis, *_ = np.argsort(chunk_shape)
        elif num_axes == 1:
            smallest_chunk_axis = 0

        if min(axis_sizes_bytes) > target_buffer_bytes:
            if num_axes > 1:
                # If the smallest full axis does not fit within the buffer size, form a square along the smallest axes
                sub_square_buffer_shape = np.array(chunk_shape)
                if min(axis_sizes_bytes) > target_buffer_bytes:
                    k1 = math.floor((target_buffer_bytes / chunk_bytes) ** 0.5)
                    for axis in [smallest_chunk_axis, second_smallest_chunk_axis]:
                        sub_square_buffer_shape[axis] = k1 * sub_square_buffer_shape[axis]
                    return tuple(sub_square_buffer_shape)
            elif num_axes == 1:
                # Handle the case where the single axis is too large to fit in the buffer
                k1 = math.floor(target_buffer_bytes / chunk_bytes)
                return (k1 * chunk_shape[0],)

        # Original one-shot estimation has good performance for certain shapes
        chunk_to_buffer_ratio = buffer_gb * 1e9 / chunk_bytes
        chunk_scaling_factor = math.floor(chunk_to_buffer_ratio ** (1 / num_axes))
        unpadded_buffer_shape = tuple(
            int(np.clip(a=int(x), a_min=chunk_shape[j], a_max=maxshape[j]))
            for j, x in enumerate(chunk_scaling_factor * np.array(chunk_shape))
        )

        unpadded_buffer_bytes = math.prod(unpadded_buffer_shape) * dtype.itemsize

        # Method that starts by filling the smallest axis completely or calculates the best partial fill
        padded_buffer_shape = np.array(chunk_shape)
        chunks_per_axis = np.ceil(maxshape / chunk_shape)
        small_axis_fill_size = chunk_bytes * min(chunks_per_axis)
        full_axes_used = np.zeros(shape=num_axes, dtype=bool)
        if small_axis_fill_size <= target_buffer_bytes:
            buffer_bytes = small_axis_fill_size
            padded_buffer_shape[smallest_chunk_axis] = maxshape[smallest_chunk_axis]
            full_axes_used[smallest_chunk_axis] = True
        for axis, chunks_on_axis in enumerate(chunks_per_axis):
            if full_axes_used[axis]:  # If the smallest axis, skip since already used
                continue
            if chunks_on_axis * buffer_bytes <= target_buffer_bytes:  # If multiple axes can be used together
                buffer_bytes *= chunks_on_axis
                padded_buffer_shape[axis] = maxshape[axis]
            else:  # Found an axis that is too large to use with the rest of the buffer; calculate how much can be used
                k3 = math.floor(target_buffer_bytes / buffer_bytes)
                padded_buffer_shape[axis] *= k3
                break
        padded_buffer_shape = tuple(int(value) for value in padded_buffer_shape)  # To avoid overflow from math.prod

        padded_buffer_bytes = math.prod(padded_buffer_shape) * dtype.itemsize

        if padded_buffer_bytes >= unpadded_buffer_bytes:
            return padded_buffer_shape
        else:
            return unpadded_buffer_shape


class SliceableDataChunkIterator(GenericDataChunkIterator):
    """
    Generic data chunk iterator that works for any memory mapped array, such as a np.memmap or h5py.Dataset object.
    """

    def __init__(self, data, **kwargs):
        self.data = data
        super().__init__(**kwargs)

    def _get_dtype(self) -> np.dtype:
        return self.data.dtype

    def _get_maxshape(self) -> tuple:
        return self.data.shape

    def _get_data(self, selection: tuple[slice]) -> np.ndarray:
        return self.data[selection]



from datetime import datetime
from typing import Optional
from warnings import warn

from pydantic import DirectoryPath, validate_call

from ..baserecordingextractorinterface import BaseRecordingExtractorInterface


class OpenEphysLegacyRecordingInterface(BaseRecordingExtractorInterface):
    """
    Primary data interface for converting legacy Open Ephys data (.continuous files).

    Uses :py:class:`~spikeinterface.extractors.OpenEphysLegacyRecordingExtractor`.
    """

    display_name = "OpenEphys Legacy Recording"
    associated_suffixes = (".continuous", ".openephys", ".xml")
    info = "Interface for converting legacy OpenEphys recording data."

    @classmethod
    def get_stream_names(cls, folder_path: DirectoryPath) -> list[str]:
        from spikeinterface.extractors import OpenEphysLegacyRecordingExtractor

        stream_names, _ = OpenEphysLegacyRecordingExtractor.get_streams(folder_path=folder_path)
        return stream_names

    @classmethod
    def get_source_schema(cls):
        """Compile input schema for the RecordingExtractor."""
        source_schema = super().get_source_schema()
        source_schema["properties"]["folder_path"][
            "description"
        ] = "Path to directory containing OpenEphys legacy files."

        return source_schema

    @validate_call
    def __init__(
        self,
        folder_path: DirectoryPath,
        stream_name: Optional[str] = None,
        block_index: Optional[int] = None,
        verbose: bool = True,
        es_key: str = "ElectricalSeries",
    ):
        """
        Initialize reading of OpenEphys legacy recording (.continuous files).

        See :py:class:`~spikeinterface.extractors.OpenEphysLegacyRecordingExtractor` for options.

        Parameters
        ----------
        folder_path : FolderPathType
            Path to OpenEphys directory.
        stream_name : str, optional
            The name of the recording stream.
        block_index : int, optional, default: None
            The index of the block to extract from the data.
        verbose : bool, default: True
        es_key : str, default: "ElectricalSeries"
        """
        available_streams = self.get_stream_names(folder_path=folder_path)
        if len(available_streams) > 1 and stream_name is None:
            raise ValueError(
                "More than one stream is detected! "
                "Please specify which stream you wish to load with the `stream_name` argument. "
                "To see what streams are available, call "
                "`OpenEphysRecordingInterface.get_stream_names(folder_path=...)`."
            )
        if stream_name is not None and stream_name not in available_streams:
            raise ValueError(
                f"The selected stream '{stream_name}' is not in the available streams '{available_streams}'!"
            )

        super().__init__(
            folder_path=folder_path, stream_name=stream_name, block_index=block_index, verbose=verbose, es_key=es_key
        )

    def get_metadata(self):
        metadata = super().get_metadata()

        neo_reader = self.recording_extractor.neo_reader
        block_annotations = neo_reader.raw_annotations.get("blocks", [])
        if block_annotations:
            segment_annotations = block_annotations[0].get("segments", [])
            if segment_annotations:
                expected_date_format = "%d-%b-%Y %H%M%S"
                date_created = segment_annotations[0]["date_created"]
                date_created = date_created.strip("'")
                extracted_date, extracted_timestamp = date_created.split(" ")
                if len(extracted_timestamp) != len("%H%M%S"):
                    warn(
                        "The timestamp for starting time from openephys metadata is ambiguous "
                        f"('{extracted_timestamp}')! Only the date will be auto-populated in metadata. "
                        "Please update the timestamp manually to record this value with the highest known "
                        "temporal resolution."
                    )
                    session_start_time = datetime.strptime(extracted_date, "%d-%b-%Y")
                else:
                    session_start_time = datetime.strptime(date_created, expected_date_format)
                metadata["NWBFile"].update(session_start_time=session_start_time)

        return metadata



from numbers import Real
from typing import Optional

import numpy as np


def calculate_regular_series_rate(series: np.ndarray, tolerance_decimals: int = 6) -> Optional[Real]:
    """Calculates the rate of a series as the difference between all consecutive points.
    If the difference between all time points are all the same value, then the value of
    rate is a scalar otherwise it is None."""
    diff_ts = np.diff(series)
    rounded_diff_ts = diff_ts.round(decimals=tolerance_decimals)
    uniq_diff_ts = np.unique(rounded_diff_ts)
    rate = 1.0 / diff_ts[0] if len(uniq_diff_ts) == 1 else None
    return rate
